import numpy as np
from scipy.optimize import least_squares
from scipy.linalg import svd


def pcscore2sphere3(n_pc, X_hat, Xs, Tan, V):
    """
    Converts principal component scores to points on the sphere.

    Usage: pcscore2sphere3(n_pc, X_hat, Xs, Tan, V)

    :param n_pc: number of principal components
    :param X_hat: (d x n) matrix of principal component scores
    :param Xs: (d x n) matrix of original data
    :param Tan: tangent space at the origin
    :param V: rotation matrix to align the tangent space with the sphere
    :return: (d x n) matrix of points on the sphere
    """
    d = Tan.shape[0]
    n = Tan.shape[1]
    W = np.zeros((d, n))
    for i in range(n):
        W[:, i] = (
            np.arccos(np.sum(Xs[i, :] * X_hat)) * Tan[:, i] / np.linalg.norm(Tan[:, i])
        )

    lam = np.zeros((n, d))
    for i in range(n):
        for j in range(n_pc):
            lam[i, j] = np.sum(W[:, i] * V[:, j])

    U = np.zeros((n, d))
    for i in range(n):
        for j in range(n_pc):
            U[i, :] = U[i, :] + lam[i, j] * V[:, j]
    S_star = np.zeros((n, n_pc + 1))
    for i in range(n):
        U_norm = np.linalg.norm(U[i, :])
        S_star[i, :] = np.insert(np.sin(U_norm) / U_norm * lam[i, 0:n_pc], 0, np.cos(U_norm))
    return S_star


def rotMat(b):
    """ "
    Returns a rotation matrix that rotates unit vector b to a

    Usage: rot = rotMat(b) returns a d x d rotation matrix that rotate
    unit vector b to the north pole (0,0,...,0,1)

    :param b: (d x 1) vector

    """
    if b.ndim == 1:
        s1 = b.shape[0]
        s2 = 1
    else:
        s1, s2 = b.shape
    d = np.maximum(s1, s2)
    b = b / np.linalg.norm(b)
    if s1 <= s2:
        b = b.T
    a = np.zeros((d, 1))
    a[d - 1] = 1
    alpha = np.arccos(a.T @ b)
    if np.abs(a.T @ b - 1) < 1e-15:
        rot = np.eye(d)
        return rot
    if np.abs(a.T @ b + 1) < 1e-15:
        rot = -1 * np.eye(d)
        return rot

    c = b - a @ (a.T @ b)
    c = c / np.linalg.norm(c)
    A = a @ c.T - c @ a.T
    rot = np.eye(d) + np.sin(alpha) * A + (np.cos(alpha) - 1) * (a @ a.T + c @ c.T)
    return rot


def LogNPd(x):
    """
    Riemannian log map at North pole of S^k

    LogNP(x) returns k x n matrix where each column is a point on tangent
       space at north pole and the input x is (k+1) x n matrix where each
       column is a point on a sphere.
    """
    d, n = x.shape
    scale = np.arccos(x[d - 1, :]) / np.sqrt(1 - x[d - 1, :] ** 2)
    scale[np.isnan(scale)] = 1
    Logpx = np.tile(scale, (d - 1, 1)) * x[0:(d - 1), :]
    return Logpx


def ExpNPd(x):
    """
    Riemannian exponential map at North pole of S^k

    ExpNP(v) returns (k+1) x n matrix where each column is a point on a
                sphere and the input v is k x n matrix where each column
                is a point on tangent  space at north pole.
    """
    d, n = x.shape
    nv = np.sqrt((x**2).sum(axis=0))
    with np.errstate(divide="ignore", invalid="ignore"):
        tmp = np.tile(np.sin(nv) / nv, (d, 1))
    Exppx = np.vstack((tmp * x, np.cos(nv)))
    tmp1 = np.zeros((d + 1, 1))
    tmp1[d] = 1
    Exppx[:, nv < 1e-16] = np.tile(tmp1, (1, (nv < 1e-16).sum()))
    return Exppx


def LMFsphereRes(center, data, greatCircle):
    """
    Auxiliary function for LMFsphereFit.
              -Calculates residuals of circle fit for the given center.
    """
    xmc = data - np.tile(center[:, np.newaxis], (1, data.shape[1]))
    di = np.sqrt(np.sum(xmc**2, axis=0))
    r = np.pi / 2
    if greatCircle == 0:
        r = di.mean()
    rr = (di - r).T

    return rr


def LMFsphereJac(center, data, greatCircle):
    """
    Auxiliary function for LMFsphereFit.
              -Calculates jacobian of circle fit for the given center.
    """
    xmc = data - np.tile(center[:, np.newaxis], (1, data.shape[1]))
    di = np.sqrt(np.sum(xmc**2, axis=0))
    di_vj = -xmc / np.tile(di, (center.shape[0], 1))
    if greatCircle:
        return di_vj.T
    else:
        r_vj = di_vj.mean(axis=1)
        return (di_vj - np.tile(r_vj[:, np.newaxis], (1, data.shape[1]))).T


def LMFsphereFit(data, initialCenter, greatCircle=True):
    """
    The least square estimates of the sphere to the data.
              The non-linear least square estimates are calculated by
              the Levenberg-Marquardt method

    center= LMFsphereFit(A) with d x n data matrix A (any d = 2, 3, ...).
    center, r = LMFsphereFit(A) with d x n data matrix gives the center and
                              the radius.
    center, r = LMFsphereFit(A,1) forces the sphere radius as 1

    @param data: points on sphere
    @param initialCenter: vector of initial guess
    @param greatCircle: if True, fit a great circle; if False, fit a sphere

    Example:
    n = 50
    theta = np.linspace(0,np.pi*1.5,n)
    data = 5*np.vstack((np.cos(theta),np.sin(theta))) + np.random.normal(size=(2,n))
    x, r = LMFsphereFit(data, np.zeros((2,1)))
    Estcirc = r*np.vstack((np.cos(theta), np.sin(theta)))+np.tile(x,(1,data.shape[1]))
    plt.scatter(data[0,:], data[1,:])
    plt.plot(x[0],x[1],'or')
    plt.plot(Estcirc[0,:],Estcirc[1,:])
    """
    result = least_squares(
        LMFsphereRes,
        initialCenter.ravel(),
        jac=LMFsphereJac,
        args=(data, greatCircle),
        method="lm",
        xtol=1e-09,
        max_nfev=1000,
    )

    center = result.x

    center = center[:, np.newaxis]

    xmc = data - np.tile(center, (1, data.shape[1]))
    di = np.sqrt(np.sum(xmc**2, axis=0))

    if greatCircle == 0:
        r = di.mean()
    else:
        r = np.pi / 2

    return center, r


def objfn(center, r, data):
    """
    the objective function that we want to minimize: sum of squared distances
    from the data to the subsphere
    """
    g = np.mean((np.arccos(center.T @ data) - r) ** 2)
    return g


def getSubSphere(data, geodesic=0):
    """
    The least square estimates of the best fitting subsphere
                to the data on the unit hyper-sphere.

    center = getSubSphere(data), with d x n data matrix with
            each column having unit length, returns the center
    center, r = getSubSphere(data), with d x n data matrix with each
               column having unit length, returns the center and the
               geodesic radius.
    center, r= getSubSphere(data, 1) forces the subsphere radius as 1

    """
    # last singular vector
    U, dd, Vh = np.linalg.svd(data)
    initialCenter = U[:, -1]

    c0 = initialCenter[:, np.newaxis]
    TOL = 1e-10
    cnt = 0
    err = 1
    d, n = data.shape
    Gnow = 1e10
    while err > TOL:
        # normalize the new candidate
        c0 /= np.linalg.norm(c0)
        # rotation matrix : c0 -> North Pole
        rot = rotMat(c0)
        # Tangent projection by Log map
        TpData = LogNPd(rot @ data)
        newCenterTp, r = LMFsphereFit(TpData, np.zeros((d - 1, 1)), geodesic)
        if r > np.pi:
            r = np.pi / 2
            U, dd, Vh = np.linalg.svd(TpData)
            newCenterTp = U[:, -1] * np.pi / 2
            newCenterTp = newCenterTp[:, np.newaxis]

        # Bring back to the sphere by Exp map
        newCenter = ExpNPd(newCenterTp)
        # rotate back the newCenter
        center = np.linalg.inv(rot) @ newCenter
        Gnext = objfn(center, r, data)
        err = np.abs(Gnow - Gnext)
        Gnow = Gnext
        c0 = center
        cnt += 1
        if cnt > 30:
            break

    i1save = {"Gnow": Gnow, "center": center, "r": r}

    # last eigenvector of covariance matrix
    K = np.cov(data)
    U, dd, Vh = np.linalg.svd(K, hermitian=True)
    initialCenter = U[:, -1]

    c0 = initialCenter[:, np.newaxis]
    TOL = 1e-10
    cnt = 0
    err = 1
    d, n = data.shape
    Gnow = 1e10
    while err > TOL:
        # normalize the new candidate
        c0 /= np.linalg.norm(c0)
        # rotation matrix : c0 -> North Pole
        rot = rotMat(c0)
        # Tangent projection by Log map
        TpData = LogNPd(rot @ data)
        newCenterTp, r = LMFsphereFit(TpData, np.zeros((d - 1, 1)), geodesic)
        if r > np.pi:
            r = np.pi / 2
            U, dd, Vh = np.linalg.svd(TpData)
            newCenterTp = U[:, -1] * np.pi / 2
            newCenterTp = newCenterTp[:, np.newaxis]

        newCenter = ExpNPd(newCenterTp)
        center = np.linalg.inv(rot) @ newCenter
        Gnext = objfn(center, r, data)
        err = np.abs(Gnow - Gnext)
        Gnow = Gnext
        c0 = center
        cnt += 1
        if cnt > 30:
            break

    if i1save["Gnow"] == np.minimum(Gnow, i1save["Gnow"]):
        center = i1save["center"]
        r = i1save["r"]

    # adjust radius
    if r > np.pi / 2:
        center *= -1
        r = np.pi - r

    return center, r


def geodmeanS1(theta):
    """
    geodesic mean of data on S^1 (Circle) by S. Lu and V. Kulkarni
               method - gives all multiples of geodesic mean set.

    @param theta: a column vector of angles

    @return:
    geodmean: geodesic mean on S^1
    geodvar: geodesic variance on S^2

    """
    n = theta.shape[0]
    meancandi = np.mod(np.mean(theta) + 2 * np.pi * np.arange(0, n) / n, 2 * np.pi)
    theta = np.mod(theta, 2 * np.pi)
    geodvar = np.zeros(n)
    for i in range(n):
        v = meancandi[i]
        tmparray = np.vstack(
            (
                (theta - v) ** 2,
                (theta - v + 2 * np.pi) ** 2,
                (v - theta + 2 * np.pi) ** 2,
            )
        )
        dist2 = tmparray.min(axis=0)
        geodvar[i] = dist2.sum()

    ind = np.argmin(geodvar)
    geodmean = np.mod(meancandi[ind], 2 * np.pi)
    geodvar = geodvar[ind] / n

    return geodmean, geodvar


def PNSs2e(spheredata, PNS):
    """
    PNS Sphere to Euclidean-type representation

    Usage: EuclidData = PNSs2e(A,PNS)

      with d x n matrix A consists of column
        vectors that are on the (d-1) sphere, and PNS is the output from
    """
    kk, n = spheredata.shape
    Res = np.zeros((kk - 1, n))
    currentSphere = spheredata.copy()

    for i in range(kk - 1):
        v = PNS["orthaxis"][i]
        r = PNS["dist"][i]
        res = np.arccos(v.T @ currentSphere) - r
        Res[i, :] = res
        NestedSphere = rotMat(v) @ currentSphere
        currentSphere = NestedSphere[0:(kk - i), :] / np.tile(
            np.sqrt(1 - NestedSphere[-1, :] ** 2), (kk - i, 1)
        )

    S1toRadian = np.arctan2(currentSphere[1, :], currentSphere[0, :])
    devS1 = np.mod(S1toRadian - PNS["orthaxis"][-1] + np.pi, 2 * np.pi) - np.pi
    Res[kk - 2, :] = devS1

    EuclidData = np.flipud(np.tile(np.flipud(PNS["radii"]), (1, n)) * Res)

    return EuclidData


def PNSe2s(resmat, PNS):
    """
    PNS coordinate transform from Euclidean-type residual matrix to Sphere

    Usage: Spheredata = PNSe2s(data,PNS)

      where 'data' is d x m data matrix in PNS coordinate system (for any
      m >= 1), 'PNS' is the structural array
    """
    dm, n = resmat.shape
    # dm is the intrinsic dimension of the sphere
    #    or the reduced sphere in HDLSS case.
    # n  is the sample size.
    NSOrthaxis = PNS["orthaxis"][:-1]
    NSOrthaxis.reverse()
    NSradius = np.flipud(PNS["dist"])
    geodmean = PNS["orthaxis"][-1]

    # standardize components
    res = resmat / np.tile(np.flipud(PNS["radii"][0:dm]), (1, n))

    # iteratviely mapping back to S^d
    # S^1 to S^2
    tmpmat = np.tile(np.sin(NSradius[0] + res[1, :]), (2, 1)) * np.vstack(
        (np.cos(geodmean + res[0, :]), np.sin(geodmean + res[0, :]))
    )
    tmpmat = np.vstack((tmpmat, np.cos(NSradius[0] + res[1, :])))
    T = rotMat(NSOrthaxis[0]).T @ tmpmat
    # S^2 to S^d
    for i in range(dm - 2):
        tmpmat = np.vstack(
            (
                np.tile(np.sin(NSradius[i + 1] + res[i + 2, :]), (3 + i, 1)) * T,
                np.cos(NSradius[i + 1] + res[i + 2, :]),
            )
        )
        T = rotMat(NSOrthaxis[i + 1]).T @ tmpmat

    if PNS["basisu"].shape[0] != 0:
        T = PNS["basisu"][:, 0:T.shape[0]] @ T

    return T


def fastPNSs2e(spheredata, PNS):
    """
    Fast PNS coordinate transform from Sphere to Euclidean-type residual matrix 

    Usage: EuclidData = fastPNSs2e(spheredata, PNS)

      where 'res' is d x m data matrix in PNS coordinate system (for any
      m >= 1), 'PNS' is the structural array
    """
    muhat = PNS["muhat"]
    U = PNS["pca"] 
    muhat = PNS["muhat"]
    n_pc = PNS["n_pc"]

    Xs = spheredata.T
    n = Xs.shape[0]
    for i in range(n):
        Xs[i, :] = Xs[i, :] / np.linalg.norm(Xs[i, :])

    muhat = np.mean(Xs, axis=0)
    muhat = muhat / np.linalg.norm(muhat)

    TT = Xs.copy()
    for i in range(n):
        TT[i, :] = Xs[i, :] - np.sum(Xs[i, :] * muhat) * muhat

    TT = TT.T

    ans = pcscore2sphere3(n_pc, muhat, Xs, TT, U)
    Xssubsphere = ans.T

    EuclidData = PNSs2e(Xssubsphere,PNS)

    return EuclidData


def fastPNSe2s(res, PNS):
    """
    Fast PNS coordinate transform from Euclidean-type residual matrix to Sphere

    Usage: Spheredata = fastPNSe2s(res, PNS)

      where 'res' is d x m data matrix in PNS coordinate system (for any
      m >= 1), 'PNS' is the structural array
    """
    GG = PNSe2s(res, PNS)
    n = GG.shape[1]
    muhat = PNS["muhat"]
    n_pc = PNS["n_pc"]
    s = np.arccos(GG[0, :])
    ones = np.ones((n, 1))
    approx1 = GG[1:(n_pc + 1), :].T @ PNS["pca"][:, 0:n_pc].T + np.diag(
        np.cos(s)
    ) @ ones @ muhat[:, np.newaxis].T / np.linalg.norm(muhat)

    return approx1


def PNSmainHDLSS(data, itype="small", a=0.05, R=100, thresh=1e-15):
    """
    Analysis of Principal Nested Spheres for data on hyperspheres

    Usage: resmat, PNS = PNSmain(data)

    @param data: (d+1) x n data matrix where each column is a unit vector.
    @param itype:  'seq_test' : (default) ordinary Principal Nested Sphere
                                with sequential tests (not implemented)
                   'small'    : Principal Nested SMALL Sphere
                   'great'    : Principal Nested GREAT Sphere (radius pi/2)
    @param alpha: 0.05 (default) : size of Type I error allowed for each test,
                 could be any number between 0 and 1.
    @param R:100 (default) : number of bootsrap samples to be evaluated for
                             the sequential test.
    @param thresh: 1e-15 (default): eigenvalue threshold

    @return:
    resmat: The commensurate residual matrix (X_PNS). Each entry in row k
            works like the kth principal component score.
    PNS: Dictionary with the following fields
        mean     : location of the PNSmean
        radii    : size (radius) of PNS
        orthoaxis: orthogonal axis 'v_i' of subspheres
        dist     : distance 'r_i' of subspheres
        pvalues  : p-values of LRT and parametric boostrap tests (if any)
        itype    : type of methods for fitting subspheres

    """
    if itype == "small":
        itype = 1
    elif itype == "great":
        itype = 2
    else:
        raise Exception("itype not implemented")

    # data on (k-1)-sphere in Real^k
    k, n = data.shape
    U, s, Vh = np.linalg.svd(data, full_matrices=False)

    maxd = np.where(s < thresh)[0]
    if maxd.shape[0] == 0 or k > n:
        maxd = np.minimum(k, n) + 1
    else:
        maxd = maxd[0]
    # dimension of subspace that contains no data
    nullspdim = k - maxd + 1

    d = k - 1  # intrinsic dimension of sphere
    print("dataset is on %d-sphere" % d)

    dm = maxd - 2  # intrinsic dimension of the smallest nested sphere that
    # contains variation. I.e. this is the dimension to which
    # can be trivially reduced.
    resmat = np.zeros((dm, n))  # d dimensional residual matrix
    # there will be dm -1 subspheres
    orthaxis = []
    dist = np.zeros((dm - 1, 1))

    if nullspdim > 0:
        print("..found null space of dimension %d to be trivally reduced" % nullspdim)
        print(".. then narrow down to %d-sphere" % dm)
        # (HDLSS case) fit nested great spheres for dimension reduction
        # where no residual is present.
        currentSphere = U[:, 0:(dm + 1)].T @ data
    else:
        print(".. Check that the following holds: %d = %d" % (d, dm))
        currentSphere = data

    pvalues = np.nan
    for i in range(dm - 1):
        # estimate the best fitting subsphere
        # with small sphere if itype = 1
        # with great sphere if itype = 2
        center, r = getSubSphere(currentSphere, itype - 1)
        res = np.arccos(center.T @ currentSphere) - r
        dist[i] = r
        # save subsphere parameters
        orthaxis.append(center)
        # save residuals
        resmat[i, :] = res
        # projection to subsphere and transformation to isomorphic
        # sphere
        NestedSphere = rotMat(center) @ currentSphere
        currentSphere = NestedSphere[0:(dm - i), :] / np.tile(
            np.sqrt(1 - NestedSphere[dm - i, :] ** 2), (dm - i, 1)
        )

    # currentSphere has intrinsic dimension 1
    # compute PNSmean and deviations.

    # parametrize 1-sphere to angles
    S1toRadian = np.arctan2(currentSphere[1, :], currentSphere[0, :])
    # geodesic mean of angles
    meantheta, vartheta = geodmeanS1(S1toRadian.T)
    orthaxis.append(meantheta)
    # save deviations from PNSmean
    resmat[dm - 1, :] = np.mod(S1toRadian - meantheta + np.pi, 2 * np.pi) - np.pi

    radii = [1]
    for i in range(dm - 1):
        radii.append(np.prod(np.sin(dist[0:i])))

    radii = np.array(radii)
    radii = radii[:, np.newaxis]
    resmat = np.flipud(np.tile(radii, (1, n)) * resmat)

    if nullspdim > 0:
        basisu = U[:, 0:(dm + 1)]
    else:
        basisu = np.empty((0))

    PNS = {
        "radii": radii,  # size (radius) of nested spheres from largest to smallest
        "orthaxis": orthaxis,  # orthogonal axis of (d-1) subspheres and the anglemean for PNSmean
        "dist": dist,  # distances for (d-1) subspheres
        "pvalues": pvalues,  # d-1 pvalues from sequential tests
        "basisu": basisu,
        "type": itype,
    }

    pnsmean = PNSe2s(np.zeros((dm, 1)), PNS)  # PNSmean of the data

    PNS["mean"] = pnsmean

    return resmat, PNS


def fastpns(x, n_pc="Full", itype="small", a=0.05, R=100, thresh=1e-15):
    """
    Fast Principal Nested Spheres (PNS) for data on hyperspheres.

    Usage: resmat, PNS = fastpns(x)

    @param x: (d+1) x n data matrix where each column is a unit vector.
    @param n_pc: number of principal components to use, or "Full" for
                 all and "Approx" for an approximate number based on
                 99% explained variance
    @param itype: 'small' or 'great' for small or great spheres.
    @param a: significance level for tests.
    @param R: number of bootstrap samples for tests.
    @param thresh: threshold for eigenvalues.

    @return:
    resmat: Commensurate residual matrix (X_PNS).
    PNS: Dictionary with PNS parameters.
    """
    n = x.shape[1]
    pdim = x.shape[0]
    if n_pc == "Full":
        n_pc = np.minimum(pdim - 1, n - 1)
    if n_pc == "Approx" or n_pc == pdim - 1:
        K = np.cov(x)
        U, s, V = svd(K)
        cumm_coef = np.cumsum(s) / np.sum(s)
        n_pc = int(np.argwhere(cumm_coef <= 0.99)[-1])

    Xs = x.T
    for i in range(n):
        Xs[i, :] = Xs[i, :] / np.linalg.norm(Xs[i, :])

    muhat = np.mean(Xs, axis=0)
    muhat = muhat / np.linalg.norm(muhat)

    TT = Xs.copy()
    for i in range(n):
        TT[i, :] = Xs[i, :] - np.sum(Xs[i, :] * muhat) * muhat

    TT = TT.T
    K = np.cov(TT)
    U, s, V = svd(K)
    pcapercent = np.sum(s[0:n_pc] ** 2 / np.sum(s**2))
    print("Initial PNS subsphere dimension: %d\n" % (n_pc + 1))
    print(
        "Percentage of variability in PNS sequence %f" % np.round(pcapercent * 100, 2)
    )

    ans = pcscore2sphere3(n_pc, muhat, Xs, TT, U)
    Xssubsphere = ans.T

    resmat, PNS = PNSmainHDLSS(Xssubsphere, itype, a, R, thresh)

    PNS["spheredata"] = Xssubsphere
    PNS["pca"] = U
    PNS["muhat"] = muhat
    PNS["n_pc"] = n_pc

    return resmat, PNS



"""
Horizontal Functional Principal Nested Spheres Analysis using SRSF

moduleauthor:: J. Derek Tucker <jdtuck@sandia.gov>

"""

import numpy as np
import fdasrsf as fs
import fdasrsf.pns as pns
import fdasrsf.utility_functions as uf
from scipy.integrate import cumulative_trapezoid
import matplotlib.pyplot as plt
import fdasrsf.plot_style as plot


class fdahpns:
    """
    This class provides horizontal fPNS using the
    SRVF framework

    Usage:  obj = fdapns(warp_data)

    :param warp_data: fdawarp class with alignment data
    :param gam_pca: warping functions principal directions
    :param psi_pca: srvf principal directions
    :param latent: latent values
    :param U: eigenvectors
    :param coef: coefficients
    :param vec: shooting vectors
    :param mu: Karcher Mean
    :param tau: principal directions

    Author :  J. D. Tucker (JDT) <jdtuck AT sandia.gov>
    Date   :  03-Apr-2025
    """

    def __init__(self, fdawarp):
        """
        Construct an instance of the fdahpns class
        :param fdawarp: fdawarp class
        """
        if fdawarp.fn.size == 0:
            raise Exception("Please align fdawarp class using srsf_align!")

        self.warp_data = fdawarp

    def calc_pns(self, var_exp=0.99, stds=np.arange(-1, 2)):
        """
        This function calculates horizontal functional principal nested
        spheres on aligned data

        :param var_exp: compute no based on value percent variance explained
                        (example: 0.95)
        :param stds: number of standard deviations along geodesic to compute
                     (default = -1,0,1)

        :rtype: fdapns object of numpy ndarray
        :return gam_pca: srsf principal directions
        :return psi_pca: functional principal directions
        :return latent: latent values
        :return coef: coefficients
        :return U: eigenvectors

        """

        gam = self.warp_data.gam
        d, n = gam.shape
        t = np.linspace(0, 1, d)
        mu, gam_mu, psi, vec = uf.SqrtMean(gam)

        radius = np.mean(np.sqrt((psi**2).sum(axis=0)))
        pnsdat = psi / np.tile(np.sqrt((psi**2).sum(axis=0)), (d, 1))

        resmat, PNS = fs.pns.fastpns(pnsdat, n_pc="Approx")

        # Proportion of variance explained
        varPNS = np.sum(np.abs(resmat) ** 2, axis=1) / n
        cumvarPNS = np.cumsum(varPNS)
        propcumPNS = cumvarPNS / cumvarPNS[-1]

        # Projection of PCs
        if (not np.any((propcumPNS <= var_exp))):
            no = 1
        else:
            no = int(np.argwhere(propcumPNS <= var_exp)[-1])+1
        if (no == 1):
            no += 1
        projPsi = np.zeros((d, stds.shape[0], no))
        projGam = np.zeros((d, stds.shape[0], no))
        for PCnum in range(no):
            std = resmat[PCnum, :].std()
            mean = resmat[PCnum, :].mean()
            dirtmp = stds * std + mean
            restmp = np.zeros((resmat.shape[0], stds.shape[0]))
            restmp[PCnum, :] = dirtmp

            PCvec = fs.pns.fastPNSe2s(restmp, PNS)
            projPsi[:, :, PCnum] = PCvec * radius
            gamt = cumulative_trapezoid(projPsi[:, :, PCnum] ** 2, t, axis=0, 
                                        initial=0)
            for j in range(stds.shape[0]):
                gamt[:, j] = (gamt[:, j] - gamt[:, j].min()) / (
                    gamt[:, j].max() - gamt[:, j].min()
                )
            projGam[:, :, PCnum] = gamt

        self.gam_pns = projGam
        self.psi_pns = projPsi
        self.cumvar = propcumPNS
        self.no = no
        self.psi = psi
        self.PNS = PNS
        self.coef = resmat
        self.radius = radius

        return

    def project(self, f):
        """
        project new data onto fPNS basis

        Usage: obj.project(f)

        :param f: numpy array (MxN) of N functions on M time

        """

        q1 = fs.f_to_srsf(f, self.warp_data.time)
        M = self.warp_data.time.shape[0]
        n = q1.shape[1]
        mq = self.warp_data.mqn
        gam = np.zeros((M, n))
        for ii in range(0, n):
            gam[:, ii] = fs.optimum_reparam(mq, self.warp_data.time, q1[:, ii])

        psi = np.zeros((M, n))
        time = np.linspace(0, 1, M)
        binsize = np.mean(np.diff(time))
        for i in range(0, n):
            psi[:, i] = np.sqrt(np.gradient(gam[:, i], binsize))

        pnsdat = psi / np.tile(np.sqrt((psi**2).sum(axis=0)), (n, 1))

        dat = self.PNS["basisu"].T @ pnsdat

        resmat = pns.fastPNSe2s(dat, self.PNS)

        self.new_coef = resmat

        return
    
    def plot(self):
        """
        plot plot elastic horizontal fPNS results

        Usage: obj.plot()
        """

        no = self.no
        TT = self.warp_data.time.shape[0]
        num_plot = int(np.ceil(no / 3))
        colors = [
            "#66C2A5",
            "#FC8D62",
            "#8DA0CB",
            "#E78AC3",
            "#A6D854",
            "#FFD92F",
            "#E5C494",
            "#B3B3B3",
        ]

        k = 0
        for ii in range(0, num_plot):
            if k > (no - 1):
                break

            fig, ax = plt.subplots(1, 3)

            for k1 in range(0, 3):
                k = k1 + (ii) * 3
                axt = ax[k1]
                if k > (no - 1):
                    break

                axt.plot(np.linspace(0, 1, TT), np.squeeze(self.gam_pns[:, :, k]))
                plt.style.use("seaborn-v0_8-colorblind")
                axt.set_title("PD %d" % (k + 1))
                axt.set_aspect("equal")

            fig.set_tight_layout(True)

        cumm_coef = 100 * self.cumvar
        idx = np.arange(0, self.cumvar.shape[0]) + 1
        plot.f_plot(idx, cumm_coef, "Coefficient Cumulative Percentage")
        plt.ylabel("Percentage")
        plt.xlabel("Index")
        plt.show()

        return


def project_pns_gam(resmat, PNS, radius, time):
   
    n = resmat.shape[1]
    d = time.shape[0]
    udir = np.eye(resmat.shape[0])
    PCvec = pns.fastPNSe2s(udir@resmat, PNS) * radius
    gam_hat = np.zeros((d, n))
    for i in range(n):
        gamt = cumulative_trapezoid(PCvec[:, i]**2, time, initial=0)
        gam_hat[:, i] = (gamt - gamt.min()) / (gamt.max() - gamt.min())
   
    return gam_hat



import numpy as np
from scipy.signal import find_peaks
from scipy.spatial.distance import pdist
from scipy.cluster.hierarchy import linkage, fcluster
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from mpl_toolkits.mplot3d import Axes3D
from matplotlib.colors import LinearSegmentedColormap


def getPPDinfo(t, Fa, lam, th):
    n_lams = lam.shape[0]

    # Compute the mean of each function in FN_temp over its rows
    FNm = np.zeros((t.shape[0], len(Fa)))
    FNm[:, 0] = Fa[0].mean(axis=1)
    
    # Find indices of local maxima in the first function's mean
    idxMaxFirst, _ = find_peaks(FNm[:, 0])

    # Initalize labels and Locations for the first function
    Labels = []
    Locs = []
    Labels.append(np.arange(1, idxMaxFirst.shape[0]+1, dtype=int))
    Locs.append(idxMaxFirst)

    # Initialize the maximum label number
    labelMax = Labels[0].max()

    # Process each function to assign labels and locate peaks
    for i in range(n_lams - 1):
        currentLabel = Labels[i]
        Labelst, labelMax = peak_successor(Fa[i], Fa[i+1], currentLabel, labelMax, 1)
        Labels.append(Labelst)

        # Find peak locations in the next function's mean
        FnmNextMean = Fa[i+1].mean(axis=1)
        idxMaxNext, _ = find_peaks(FnmNextMean)
        Locs.append(idxMaxNext)

        # Update the mean function
        FNm[:, i+1] = FnmNextMean
    
    IndicatorMatrix, Heights, Heights2 = PreprocessingForPPD(t, lam, Labels, Locs, labelMax, FNm, th)

    if np.all(np.isnan(Heights2)):
        NameError("All peaks are ignored. A smaller threshold is required.")

    return (IndicatorMatrix, Heights, Locs, Labels, FNm)


def peak_successor(f1, f2, labels1, labelMax, smooth_parameter):
    # Combine f1 and f2 into a 3D array and compute the mean across
    # the second dimension
    F = np.concatenate((f1[:,:,np.newaxis], f2[:,:,np.newaxis]), axis=2)
    fm = F.mean(axis=1)

    # compute peak ranges and labes for fm[:,0]
    ranges, idx_max1 = computePeakRanges(fm[:, 0])

    # compute indices of local maxima in fm[:,1]
    idx_max2, _ = find_peaks(fm[:,1])

    if idx_max1.size==0:
        labels2 = labelMax + np.arange(1, idx_max2.shape[0]+1, dtype=int)
    else:
        # assign labels to peaks in fm[:,1] based on matching ranges in fm[:,0]
        labels2 = assignLabelsToPeaks(idx_max2, ranges, idx_max1, labels1)

        # ensure no overlapping labels in lables2
        labels2 = resolveOverlappingLabels(labels2, idx_max1, idx_max2, labels1)

        # assign new labels to unmatched peaks in fm[:,1]
        unmatched = labels2 == 0
        labels2[unmatched] = labelMax + np.arange(1,np.sum(unmatched)+1, dtype=int)
        labelMax = labelMax + sum(unmatched)
    
    return(labels2, labelMax)


def computePeakRanges(data):
    # computes peak ranges defined by adjacent minima in the data
    idx_max, _ = find_peaks(data)
    if idx_max.size == 0:
        idx_max = []
        ranges = []
    
    idx_min, _ = find_peaks(-1*data)
    idx_min = np.insert(idx_min, 0, 0)
    idx_min = np.insert(idx_min, 0, data.shape[0])
    idx_min = np.unique(idx_min)
    ranges = np.zeros((idx_max.shape[0], 2))
    for i in range(idx_max.shape[0]):
        ranges[i,0] = np.max(idx_min[idx_min<idx_max[i]])
        ranges[i,1] = np.min(idx_min[idx_min>idx_max[i]])
    # remove degenerate ranges
    idx = ranges[:,0] == ranges[:,1]
    ranges = np.delete(ranges, idx, 0)
    return(ranges, idx_max)


def assignLabelsToPeaks(idx_max2, ranges, idx_max1, labels1):
    # assigns labels to peaks in idx_max2 based on matching ranges in idx_max1
    labels = np.zeros(idx_max2.shape[0], dtype=int)
    for i in range(idx_max2.shape[0]):
        # find the range in fm[:,0] that contains the current peak in fm[:,1]
        in_range = np.logical_and(idx_max2[i] >= ranges[:,0], idx_max2[i] <= ranges[:,1])
        matching_ranges = np.argwhere(in_range)

        if matching_ranges.size != 0:
            # choose the closest peak if multiple ranges match
            if matching_ranges.size > 1:
                tmp = np.abs(idx_max1[matching_ranges] - idx_max2[i])
                closest_idx = tmp.argmin()
                matching_range = matching_ranges[closest_idx]
            else:
                matching_range = matching_ranges.copy()
            
            labels[i] = labels1[matching_range]

    return labels


def resolveOverlappingLabels(labels, idx_max1, idx_max2, labels1):
    # ensures no overlapping labels in the assigned labels
    unique_labels = np.unique(labels[labels > 0])
    for label in unique_labels:
        duplicates = np.argwhere(labels==label)
        if duplicates.size > 1:
            # keep the closest peak and reset others
            distances = np.abs(idx_max1[labels1==label] - idx_max2[duplicates])
            min_idx = distances.arg_min()
            duplicates = np.delete(duplicates, min_idx, 0)
            labels[duplicates] = 0
    return labels


def PreprocessingForPPD(t, lam, Labels, Locs, labelMax, FNm, th):
    K = lam.shape[0]
    # initialize output matrices
    IndicatorMatrix = np.full((K, labelMax), np.nan) 
    curvatures = np.zeros((K, labelMax))
    Heights = np.full((K, labelMax), np.nan)

    # assume t is uniformly spaced; compute time step
    dx = t[1] - t[0]

    # process each function
    for i in range(K):
        # extract the function values for the current parameter lambda
        fnm = FNm[:, i]

        # compute negative curvature (sec)
        negCurvature = -1*np.gradient(np.gradient(fnm,dx), dx)

        # Ensure non-negative curvature values
        negCurvature[negCurvature < 0] = 0

        # normalize negative curvature to [0,1] if possible 
        maxNegCurvature = negCurvature.max()
        if maxNegCurvature > 0:
            negCurvature = negCurvature / negCurvature
        
        # retrieve peak locations and labels for the current function
        locsCurrent = Locs[i]
        labelsCurrent = Labels[i]

        # select negative curvature values at specified peak locations
        negCurvSelected = negCurvature[locsCurrent]

        # update curvatures and heights matrices at the appropriate labels
        curvatures[i, labelsCurrent-1] = negCurvSelected
        Heights[i, labelsCurrent-1] = fnm[locsCurrent]

        # apply threshold to select significant peaks based on curvature
        significantLabels = labelsCurrent[negCurvSelected >= th]

        # update the indicator matrix for signficiant peaks
        IndicatorMatrix[i, significantLabels-1] = 1
    
    # compute heighs2 by multiply heighs with the indicator matrix 
    Heights2 = IndicatorMatrix * Heights

    return(IndicatorMatrix, Heights, Heights2)


def getPersistentPeaks(IndicatorMatrix):

    if IndicatorMatrix.size == 0:
        Clt2 = []
        return Clt2
    
    # count the number of ones (occurences) for each peak (ignore nans)
    occurrenceCounts = np.nansum(IndicatorMatrix, 1)

    data = np.append(occurrenceCounts, 0)
    if np.all(data == 0) or np.unique(occurrenceCounts).size == 1:
        Clt2 = []
        return Clt2
    
    # compute pairwise distances between observations
    pairwiseDistances = pdist(data, metric='euclidean')
    Y = linkage(pairwiseDistances, 'ward')

    # Cluster the data into the specified number of clusters
    clusterAssignments = fcluster(Y, 2, 'maxclust')
    referenceCluster = clusterAssignments[-1]
    clusterAssignments = clusterAssignments[:-1]

    # identify indices wehre cluster assignmetns differ from the reference
    Clt2 = np.argwhere(clusterAssignments != referenceCluster)

    return Clt2


def drawPPDBarChart(IndicatorMatrix, Heights, lam, idx_opt):
    lam_diff = lam[1] - lam[0]
    len_lam, labelMax = IndicatorMatrix.shape

    fig, ax = plt.subplots()

    for i in range(len_lam):
        label_all_peaks = np.argwhere(np.logical_not(np.isnan(Heights[i,:])))
        label_persistent_peaks = np.argwhere(np.logical_not(np.isnan(IndicatorMatrix[i,:])))

        if label_all_peaks.sum() == 0:
            continue

        for j in range(label_all_peaks.shape[0]):
            x = lam[i]
            y = label_all_peaks[j,0]

            if y in label_persistent_peaks:
                rect = Rectangle((x, y), lam_diff, 1, facecolor='black', edgecolor='none')
            else:
                rect = Rectangle((x, y), lam_diff, 0.5, 1, facecolor='grey', edgecolor='none')
            
            ax.add_patch(rect)
    
    for j in range(labelMax):
        plt.hlines(y=j+0.5, xmin=plt.xlim()[0], xmax=plt.xlim()[1])
    
    plt.axvline(x=lam[idx_opt], color="red", linestyle='dashed', linewidth=2)

    plt.xlim((lam[0], lam[-1]))
    plt.ylim((0.5, labelMax))

    plt.yticks(np.arange(1,labelMax+1))

    plt.xlabel('$\\lambda$')
    plt.ylabel('Peak Index')


def drawPPDSurface(t,lam,FNm,Heights,Locs,IndicatorMatrix,Labels,idx_opt):
    parula_colors = [
        (0.24, 0.15, 0.66),  # Dark blue
        (0.26, 0.44, 0.82),  # Medium blue
        (0.28, 0.70, 0.81),  # Cyan
        (0.60, 0.80, 0.44),  # Greenish-yellow
        (0.99, 0.81, 0.19)   # Yellow
    ]

    # Create the custom colormap
    parula_cmap = LinearSegmentedColormap.from_list('parula_custom', parula_colors)
    
    n_lams = lam.shape[0]
    labelMax = IndicatorMatrix.shape[1]

    LocationMatrix_full = np.full((n_lams, labelMax), np.nan, dtype=int)
    for i in range(n_lams):
        LocationMatrix_full[i, Labels[i]-1] = Locs[i]
    
    LocationMatrix_sig = LocationMatrix_full * IndicatorMatrix
    LocationMatrix_sig = LocationMatrix_sig.astype(int)
    HeighMatrix_full = Heights.copy()
    HeightMatrix_sig = HeighMatrix_full * IndicatorMatrix

    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    X, Y = np.meshgrid(t, lam)
    ax.plot_surface(X, Y, FNm.T, cmap=parula_cmap, linewidth=0.5)

    plt.xlim(t[0], t[-1])

    ax.plot(t, lam[idx_opt]*np.ones(t.shape[0]), FNm[:, idx_opt], linestyle='dashed', linewidth=2, color='magenta')
    # loop through each label
    for j in range(labelMax):

        # find non-nan indices for full location matrix and plot
        idx_full = np.argwhere(np.logical_not(np.isnan(LocationMatrix_full[:, j]))).ravel()
        ax.plot(t[LocationMatrix_full[idx_full, j]], lam[idx_full], HeighMatrix_full[idx_full, j], marker='o', color='black', linewidth=1.5, markersize=5, mfc='black')

        # find non-nan indices for significant location matrix and plot
        idx_sig = np.argwhere(np.logical_not(np.isnan(LocationMatrix_sig[:, j]))).ravel()
        ax.plot(t[LocationMatrix_sig[:, j]], lam[idx_sig], HeightMatrix_sig[idx_sig, j], linestyle='dashed', linewidth=2, color='magenta', markersize=6)
    
    plt.xlabel('$t$')
    plt.ylabel('$\\lambda$')
    ax.set_zlabel('$\\hat{g}_\\lambda(t)$')

    plt.grid(True)
